#include <stdio.h>
#include <dos.h>
#include <string.h>
#include <stdlib.h>
#include <conio.h>

#ifndef __BOOL_TYPE__
#define __BOOL_TYPE__

typedef enum{
  TRUE  = 1,
  FALSE = 0
} boolean;

typedef unsigned char byte;
typedef unsigned int word;
typedef unsigned long int dword;

#endif

struct PCX_header{
  byte creator;  //always 0
  byte version;  //always 5
  byte encoding; //1 = RLE
  byte bpp;      //bits per pixel; should be 8 (256 = 2^8)
  word x0, y0;   //initial x, y coords
  word xlen, ylen;  //width and height
  word hres;     //horizontal resoution
  word vres;     //vertical resolution
  byte egapal[48];  //16 color palette...unused here
  byte reserved; //for new PCX versions
  byte color_planes;  //color planes
  word bpl;      //bytes per line of pic
  word paltype;  //palette type (2 for us)
  byte filler[54];   //pcx spec says these should all be 0.
};

struct D13_fs_bmp{
  byte pal[768];
};

boolean D13_test_filename(char* filename);
boolean D13_load_pcx(char* filename, D13_fs_bmp& bmpinfo);
void D13_load_pcx_pal(char* filename, D13_fs_bmp& bmpinfo);
int D13_getcurmode(void);
void D13_setgmode(word mode);
boolean D13_palette_rotl(void* D13_CURRENT_PALETTE, byte first,
                         byte last, byte number);
void D13_setpal(void* palette);
void wordswap(word& a, word& b);
void sort_frequencies(word* frequency, word* tagarray);

int encput(unsigned char byt, unsigned char cnt, FILE* fid);
int encLine(unsigned char* inBuff, int inLen, FILE* fp);
boolean snag_pcx_header(char* filename, PCX_header& header);


int D13_getcurmode(void){
  int oldmode = *(int*)MK_FP(0x40,0x49);
  return oldmode;
}

void D13_setgmode(word mode){
                asm mov     ax, mode;
                asm mov     ah, 0;
                asm int     10h;
}

/******************************************************************/
//      boolean D13_test_filename(char* filename);
//      Tests to see if the specified file is on the disk.
//      Will ARF if removable media is not present.
//
//      Takes   :  Pointer to filename
//      Modifies:  Nothing
//      Returns :  TRUE if file exists, FALSE if not.
//      Used By :  User
//
/******************************************************************/

boolean D13_test_filename(char* filename){
  FILE *testing;
  testing = fopen(filename, "r");
  if (testing == NULL){
    return FALSE;
  }
  fclose(testing);
  return TRUE;
}


/******************************************************************/
//      boolean D13_load_pcx(char* filename, D13_fs_bmp& bmpinfo);
//      Loads a 320x200x256 PCX file & palette into buffer pointed
//      at by D13_PUTPOINTER and palette in bmpinfo.
//
//      Takes   :  PCX filename, dummy fullscreen bitmap structure
//      Modifies:  Bitmap structure
//      Returns :  TRUE if all went well, FALSE if not.
//      Used By :  User
//
/******************************************************************/
boolean D13_load_pcx(char* filename, D13_fs_bmp& bmpinfo){
  if (!D13_test_filename(filename)) return FALSE;
  void* D13_PUTPOINTER = MK_FP(0xA000, 0x0000);
  FILE* pcxfile = fopen(filename, "rb");
  const int BYTEMODE=0, RUNMODE=1;
  int mode=BYTEMODE;
  static unsigned char outbyte,bytecount;
  fseek(pcxfile,128L,SEEK_SET);
  for (long i=0; i<64000; i++) {
    if (mode==BYTEMODE) {
      outbyte=(unsigned char)fgetc(pcxfile);
      if (outbyte>0xbf) {
        bytecount = (int)((int)outbyte & 0x3f);
        outbyte=(unsigned char)getc(pcxfile);
        if (--bytecount > 0) mode = RUNMODE;
      }
    }
    else if (--bytecount == 0) mode=BYTEMODE;
    *((byte*)(D13_PUTPOINTER)+i)=outbyte;
  }
  fclose(pcxfile);
  D13_load_pcx_pal(filename, bmpinfo);
  return TRUE;
}

/******************************************************************/
//      boolean D13_load_pcx_pal(char* filename, D13_fs_bmp& bmpinfo);
//      Loads a 320x200x256 PCX palette into buffer pointed
//      to by bmpinfo.
//
//      Takes   :  PCX filename, dummy fullscreen bitmap structure
//      Modifies:  Bitmap structure
//      Returns :  TRUE if all went well, FALSE if not.
//      Used By :  PCX routines
//
/******************************************************************/
void D13_load_pcx_pal(char* filename, D13_fs_bmp& bmpinfo){
  FILE* pcxfile = fopen(filename, "rb");
  fseek(pcxfile,-768L,SEEK_END);
  for (int i=0; i<256; i++)
    for (int j=0; j<3; j++) bmpinfo.pal[i*3+j]=fgetc(pcxfile)>>2;
  fclose(pcxfile);
}

boolean snag_pcx_header(char* filename, PCX_header& header){
  //this function snags a PCX header from a file.  Because we're only
  //changing the picture and the palette (not horiz. res, byte planes,
  //etc., we can use the same header!)
  void* phv = &header;
  byte* ph = (byte*)phv;
  if (!D13_test_filename(filename)) return FALSE;
  FILE* pcxfile = fopen(filename, "rb");
  fread(ph, 128, 1, pcxfile);
  fclose(pcxfile);
  return TRUE;
}

void D13_setpal(void* palette){
                asm{
                push    si
                push    ds
                push    di

                lds     si, palette
                mov     al, 0
                mov     dx, 3c8h
                out     dx, al
                mov     cx, 768
                mov     dx, 3c9h
                rep     outsb

                pop     di
                pop     ds
                pop     si
                }
}

/******************************************************************/
//      void D13_palette_rotl(byte first, byte last, byte number);
//      Rotates a segment of the current palette defined by 'first'
//      and 'last' leftwards (with wrapping), 'number' colors.
//
//      Takes   :  First byte in segment, last byte in segment,
//                 number to rotate.
//      Modifies:  Nothing
//      Returns :  TRUE if all went well, FALSE if file not found.
//      Used By :  User
//
/******************************************************************/
boolean D13_palette_rotl(byte* D13_CURRENT_PALETTE, byte first,
                         byte last, byte number){
  word f = first  * 3;
  word l = last   * 3;
  word n = number * 3;

  byte* holding = new byte[n+3];
  if (holding == NULL) return FALSE;
  word len = l - f + n;
  //memmove((void*)holding, &D13_CURRENT_PALETTE[f], n);
  int i;
  for (i = 0; i < n+1; i++) holding[i] = 0;
  memmove((void*)&D13_CURRENT_PALETTE[f],
          (void*)&D13_CURRENT_PALETTE[f+n], len);
  memmove((void*)&D13_CURRENT_PALETTE[l-n+3], (void*)holding, n);
  delete holding;
  D13_setpal(D13_CURRENT_PALETTE);
  return TRUE;
}

void wordswap(word& a, word& b){
  word temp;
  temp = a;
  a = b;
  b = temp;
}

void intswap(int& a, int& b){
  int temp;
  temp = a;
  a = b;
  b = temp;
}

void sort_frequencies(word* frequency, word* tagarray){
  long i, j;
  word least, index;
  least = frequency[0];
  index = 0;
  for (i = 0; i < 256; i++){
    least = frequency[i];
    index = i;
    for (j = i; j < 256; j++){
      if (frequency[j] <= least){
        least = frequency[j];
        index = j;
      }
    }
    wordswap(frequency[i], frequency[index]);
    wordswap(tagarray[i], tagarray[index]);
  }
}

void sort_integers(int* arr, int* tagarray, int num){
  long i, j;
  word least, index;
  least = arr[0];
  index = 0;
  for (i = 0; i < num; i++){
    least = arr[i];
    index = i;
    for (j = i; j < num; j++){
      if (arr[j] <= least){
        least = arr[j];
        index = j;
      }
    }
    intswap(arr[i], arr[index]);
    if (tagarray != NULL) intswap(tagarray[i], tagarray[index]);
  }
}

void replace_color(byte* screen, byte toreplace, byte replacement){
  long i;
  for (i = 0; i <= 64000L; i++){
    if (screen[i] == toreplace) screen[i] = replacement;
  }
}

void get_color(byte* pal, int entry, int& red, int& green, int& blue){
  red = pal[entry * 3];
  green = pal[(entry * 3) + 1];
  blue = pal[(entry * 3) + 2];
}

void delete_pal_entry(byte* palette, int entry){
  D13_palette_rotl(palette, entry, 255, 1);
  byte* screen = (byte*)MK_FP(0xA000, 0x0000);
  long i;
  int pixel;
  byte foo;
  for (i = 0; i <= 64000L; i++){
    if (screen[i] > entry) screen[i]--;
  }
}

int find_closest_color(byte* pal, int entry, int acceptable_deviation,
                       int highest){
  byte intensity_difference[768];
  long i;
  int esplit[3];
  int csplit[3];
  get_color(pal, entry, esplit[0], esplit[1], esplit[2]);
  int idtemp;
  for (i = 0; i < 256; i++){
    get_color(pal, i, csplit[0], csplit[1], csplit[2]);
    idtemp = abs(esplit[0] - csplit[0]);
    intensity_difference[(i * 3) + 0] = idtemp;
    idtemp = abs(esplit[1] - csplit[1]);
    intensity_difference[(i * 3) + 1] = idtemp;
    idtemp = abs(esplit[2] - csplit[2]);
    intensity_difference[(i * 3) + 2] = idtemp;
  }
  int ei[3]; //highest, medium, lowest intensity of entry
  int em[3]; //intensity modifiers;
  int etag[3] = {0, 1, 2};
  sort_integers(esplit, etag, 3);
  ei[2] = etag[0];
  ei[1] = etag[1];
  ei[0] = etag[2];
  em[0] = 3;
  em[1] = 2;
  em[2] = 1;
  if (ei[0] < ei[1] + 64) em[0] = 2;
  if (ei[1] < ei[2] + 64) em[2] = 2;

  long current_deviation;
  get_color(intensity_difference, 0, csplit[0], csplit[1], csplit[2]);
  long least_deviation = (csplit[0] * em[0])
                        +(csplit[1] * em[1])
                        +(csplit[2] * em[2]);
  int index_least_deviation = 0;
  for (i = 1; i < highest; i++){
    if (i != entry){
      get_color(intensity_difference, i, csplit[0], csplit[1], csplit[2]);
      current_deviation = (csplit[0] * em[0])
                         +(csplit[1] * em[1])
                         +(csplit[2] * em[2]);
      if ((current_deviation) < least_deviation){
        least_deviation = current_deviation;
        index_least_deviation = i;
      }
    }
  }
  if ((least_deviation >> 1) > acceptable_deviation) return -1;
  return index_least_deviation;
}

//The following is a set of C subroutines to write data to a .PCX file.

/* Subroutine for writing an encoded byte pair (or single byte if it
doesn't encode) to a file. It returns the count of bytes written, 0 if error */

int encput(unsigned char byt, unsigned char cnt, FILE* fid){
  if (cnt) {
    if ((cnt == 1) && (0xC0 != (0xC0 & byt))){
      if (EOF == putc((int )byt, fid))
        return(0);     /* disk write error (probably full) */
        return(1);
      }
      else{
      if (EOF == putc((int )0xC0 | cnt, fid))
        return (0);      /* disk write error */
      if (EOF == putc((int )byt, fid))  return (0); /* disk write error */
      return (2);
    }
  }
  return (0);
}

/* This subroutine encodes one scanline and writes it to a file.
It returns number of bytes written into outBuff, 0 if failed. */

int encLine(unsigned char* inBuff, int inLen, FILE* fp){
  unsigned char thisp, last;
  int srcIndex, i;
  register int total;
  register unsigned char runCount;     /* max single runlength is 63 */
  total = 0;
  runCount = 1;
  last = *(inBuff);

/* Find the pixel dimensions of the image by calculating 
[XSIZE = Xmax - Xmin + 1] and [YSIZE = Ymax - Ymin + 1].  
Then calculate how many bytes are in a "run" */

  for (srcIndex = 1; srcIndex < inLen; srcIndex++){
    thisp = *(++inBuff);
    if (thisp == last){     /* There is a "run" in the data, encode it */
      runCount++;
      if (runCount == 63){
        if (! (i = encput(last, runCount, fp))) return (0);
          total += i;
          runCount = 0;
        }
      }
      else{                /* No "run"  -  thisp != last */
        if (runCount){
          if (! (i = encput(last, runCount, fp))) return(0);
            total += i;
          }
          last = thisp;
          runCount = 1;
    }
  }        /* endloop */
  if (runCount){        /* finish up */
    if (! (i = encput(last, runCount, fp))) return (0);
      return (total + i);
    }
   return (total);
}



int main(int argc, char* argv[]){
  printf("\nArchon Software's Palette Reducer V1.0\n");
  printf("(C) 1995 Eric M. Dashofy\nAll Rights Reserved\n");
  if (argc != 4){
    printf("\nError!  Wrong number of arguments!\n");
    printf("Usage is as follows:\n\n");
    printf("PALRED [infile] [colors] [acceptable deviation]\n");
    printf("\nWhere:\n [infile] is a .PCX file.\n");
    printf(" [colors] is the number of colors to reduce to.\n");
    printf(" [acceptable deviation] is the acceptable color deviation.\n");
    return 1;
  }
  int oldmode = D13_getcurmode();
  int reduceto = atoi(argv[2]);
  if ((reduceto < 0) || (reduceto > 254)){
    printf("Error:  Number of colors must be between 0 & 254.\n");
    return 1;
  }
  int acceptdev = atoi(argv[3]);
  if ((acceptdev < 0) || (acceptdev > 1000)){
    printf("Error:  Acceptable deviation must be between 0 & 1000.\n");
    return 1;
  }
  D13_setgmode(0x13);
  D13_fs_bmp pcxfile;
  char pcxfilename[20];
  strcpy (pcxfilename, argv[1]);
  if (!D13_load_pcx(pcxfilename, pcxfile)){
    D13_setgmode(oldmode);
    printf("Invalid PCX filename!  File: %s\n", pcxfilename);
    printf("Usage is as follows:\n\n");
    printf("PALRED [infile] [colors] [acceptable deviation]\n");
    printf("\nWhere:\n [infile] is a .PCX file.\n");
    printf(" [colors] is the number of colors to reduce to.\n");
    printf(" [acceptable deviation] is the acceptable color deviation.\n");
    return 1;
  }
  PCX_header hdr;
  snag_pcx_header(pcxfilename, hdr);

  D13_setpal(pcxfile.pal);
  byte* screen = (byte*)MK_FP(0xA000, 0x0000);
  word frequency[256];
  word tagarray[256];
  long i;
  //next loops count the number of pixels of each color present to find
  //the colors which will do the least amount of damage if changed.
  for (i = 0; i < 255; i++){
    frequency[i] = 0;
    tagarray[i] = i;
  }
  frequency[255] = 0;
  tagarray[255] = 255;
  for (i = 0; i < 64000L; i++) frequency[screen[i]]++;
  sort_frequencies(frequency, tagarray);

  /* Start Group:  Remove colors that have 0 pixels assigned to them */
  int needed_to_dump = 256 - reduceto;
  int bonus = 0;
  int highest_true_entry = 255;
  i = 0;
  long j;
  int zeroes = 0;
  while(frequency[i] == 0){
    delete_pal_entry(pcxfile.pal, tagarray[i]);
    for (j = 0; j < 255; j++) if (tagarray[j] > tagarray[i]) tagarray[j]--;
    highest_true_entry--;
    if (needed_to_dump > 0) needed_to_dump--;
    else bonus++;
    zeroes++;
    i++;
  }
  if (needed_to_dump == 0){
    getch();
    D13_setgmode(oldmode);
    printf("Palette Reducer Finished!\n");
    printf("Your .PCX file's palette contained unused colors.\n");
    printf("These colors were eliminated and the %i colors\n", 256-reduceto);
    printf("you requested were converted.  Furthermore, %i\n", bonus);
    printf("additional colors were also freed.\n\n");
    return 0;
  }
  /* End Group:    Remove colors that have 0 pixels assigned to them */

  int ad;
  int closest_color;
  int number_quantized = 0;
  int number_out_of_line = 0;
  boolean outofline;
  while (needed_to_dump > 0){
    for (i = 0; i < 255; i++){
      frequency[i] = 0;
      tagarray[i] = i;
    }
    frequency[255] = 0;
    tagarray[255] = 255;
    for (i = 0; i < 64000L; i++) frequency[screen[i]]++;
    sort_frequencies(frequency, tagarray);
    i = 0;
    while(frequency[i] == 0) i++;
    closest_color = -1;
    outofline = FALSE;
    ad = acceptdev;
    while (closest_color == -1){
      closest_color = find_closest_color(pcxfile.pal, tagarray[i], ad,
                                         highest_true_entry);
      if (closest_color == -1) outofline = TRUE;
      ad+=3;
    }
    if (outofline) number_out_of_line++;
    replace_color(screen, tagarray[i], closest_color);
    delete_pal_entry(pcxfile.pal, tagarray[i]);
    for (j = 0; j < 255; j++) if (tagarray[j] > tagarray[i]) tagarray[j]--;
    sort_frequencies(frequency, tagarray);
    highest_true_entry--;
    number_quantized++;
    needed_to_dump--;
  }
  FILE* outfile = fopen("PALRED.PCX", "wb");
  fwrite(&hdr, 128, 1, outfile);
  byte* scanlinepointer;
  for (i = 0; i < 200; i++){
    scanlinepointer = (byte*)MK_FP(0xA000, 320*i);
    encLine(scanlinepointer, 320, outfile);
  }
  fputc(12, outfile);  //signal palette start
  for (i = 0; i < 768; i++) fputc(pcxfile.pal[i] << 2, outfile);
  fclose(outfile);
  D13_setgmode(oldmode);
  printf("Palette Reducer Finished!\n");
  printf("Your .PCX file's palette was reduced as requested.\n");
  printf("Of the 256 colors in the original palette:\n");
  printf("%i were unused and therefore eliminated.\n", zeroes);
  printf("%i needed to be quantized to reduce palette.\n", number_quantized);
  printf("%i were out of the acceptable deviation.\n", number_out_of_line);
  printf("%i remain in the palette.\n", highest_true_entry + 1);
  return 0;
}

